#pragma once

#include <array>
#include <functional>
#include <memory>
#include <string>
#include <unordered_map>
#include <vector>

#include "tf_data_types.h"

namespace Trueface {
class SDK;

/**
 * Class which is used to manage a face chip image.
 * A face chip represents and aligned and cropped face image.
 * A Facechip can be in CPU memory or GPU memory.
 */
class Facechip {
public:
    /**
     * Default constructor.
     */
    Facechip() = default;

    /**
     * Destructor frees allocated memory.
     */
    ~Facechip();
#if defined(TARGET_OS_IPHONE) || defined(TARGET_IPHONE_SIMULATOR)
    void destroy();
#endif

    /**
     * Save the face chip to disk.
     * @param[in] filepath The filepath where the face chip should be saved, including the image
     * extension.
     */
    void saveImage(const std::string &filepath);

    /**
     * Load the face chip from disk.
     * @param[in] filepath The filepath of the facechip to load.
     * @param[in] gpuMemory Read the image into GPU memory. This should be set to true when running
     * GPU inference.
     * @param[in] gpuIndex The GPU index.
     */
    ErrorCode loadImage(const std::string &filepath, bool gpuMemory = false, int gpuIndex = 0);

    /**
     * Get the face chip height in pixels.
     * @return Returns the face chip height in pixels.
     */
    unsigned int getHeight() const;

    /**
     * Get the face chip width in pixels.
     * @return Returns the face chip width in pixels.
     */
    unsigned int getWidth() const;

    /**
     * Get the GPU stride for GPU images.
     * @return Returns the GPU stride for the Facechip.
     */
    unsigned int getStride() const;

    /**
     * Is the face chip in GPU memory.
     * @return Returns true if the face chip is in GPU memory.
     */
    bool isGPUImage() const;

    /**
     * Get the decoded image buffer.
     * @return Returns the decoded image buffer.
     */
    uint8_t *getData() const;

private:
    unsigned int m_width = 0;
    unsigned int m_height = 0;
    unsigned int m_stride = 0;
    bool m_GPUImage = false;
    uint8_t *m_data = nullptr;
    friend class SDK;
};

/*
 * Shared pointer of type Facechip.
 */
using TFFacechip = std::shared_ptr<Facechip>;

/**
 * Class which is used to manage a decoded image which is in the format expected by inference
 * functions. Should populate Image using the preprocessImage functions. An Image can be in CPU
 * memory or GPU memory.
 */
class Image {
public:
    /**
     * Default constructor.
     */
    Image() = default;
    /**
     * Destructor frees allocated memory.
     */
    ~Image();
#if defined(TARGET_OS_IPHONE) || defined(TARGET_IPHONE_SIMULATOR)
    void destroy();
#endif

    /**
     * Rotate the image.
     *
     * @param[in] rotateFlag the RotateFlags.
     */
    void rotate(RotateFlags rotateFlag);

    /**
     * Resize the image to the target size.
     * Adds padding to the bottom and right if required in order to maintain aspect ratio.
     * @param[in] width The target width.
     * @param[in] height The target height.
     */
    void resize(unsigned int width, unsigned int height);

    /**
     * Save the preprocessed image to disk.
     * @param[in] filepath The filepath where the image should be saved, including the image
     * extension.
     */
    void saveImage(const std::string &filepath);

    /**
     * Get the image height in pixels.
     * @return Returns the image height in pixels.
     */
    unsigned int getHeight() const;

    /**
     * Get the image width in pixels.
     * @return Returns the image width in pixels.
     */
    unsigned int getWidth() const;

    /**
     * Get the number of channels for the image.
     * @return Returns the number of channels.
     */
    unsigned int getChannels() const;

    /**
     * Get the GPU stride for GPU images.
     * @return Returns the GPU stride for the image.
     */
    unsigned int getStride() const;

    /**
     * Is the image in in GPU memory.
     * @return Returns true if the image is in GPU memory.
     */
    bool isGPUImage() const;

    /**
     * Get the decoded image buffer.
     * @return Returns the decoded image buffer.
     */
    uint8_t *getData() const;

private:
    unsigned int m_height = 0;
    unsigned int m_width = 0;
    unsigned int m_channels = 0;
    unsigned int m_stride = 0;
    bool m_GPUImage = false;
    uint8_t *m_data = nullptr;
    friend class SDK;
    friend class Impl;
};

/*
 * Shared pointer of type Image.
 */
using TFImage = std::shared_ptr<Image>;

class SDK {
public:
    /**
     * Initialize the SDK using default \ref ConfigurationOptions.
     *
     * @sa ConfigurationOptions
     */
    SDK();

    /**
     * Do not allow copy.
     */
    SDK(const SDK &other) = delete;

    /**
     * Do not allow assignment operator.
     */
    SDK &operator=(const SDK &other) = delete;

    /**
     * Move constructor.
     * If other is to be used after move, must be re-initialized.
     */
    SDK(SDK &&other) noexcept;

    /**
     * Move assignment operator.
     * If other is to be used after move, must be re-initialized.
     */
    SDK &operator=(SDK &&other) noexcept;

    /**
     * Initialize the SDK using custom \ref ConfigurationOptions.
     *
     * @param[in] options custom configuration options.
     */
    SDK(const ConfigurationOptions &options);

    ~SDK();

    /**
     * Uninitialize specific module.
     *
     * @param[in] uninitializeModule module to clear.
     */
    void uninitializeModule(UninitializeModule uninitializeModule);

    /**
     *  Sets and validates the given license token. Need to call this method
     *  before being able to use the SDK.
     *
     * @param[in] token The license token (if you do not have this talk to
     *        support@pangiam.com).
     * @return Whether the given license token is valid.
     *
     * @sa isLicensed()
     */
    bool setLicense(std::string token);
#if defined(TARGET_OS_IPHONE) || defined(TARGET_IPHONE_SIMULATOR)
    bool setLicense(std::string token, std::string package_id);
#endif
    /**
     * Checks whether the given license token is valid and you can use
     * the SDK.
     *
     * @param[out] daysRemaining The number of remaining days that the token is valid for.
     * @return Whether the given license token is valid.
     *
     * @sa setLicense()
     */
    bool isLicensed(unsigned int &daysRemaining);

    /**
     * Gets the version-build number of the SDK.
     *
     * @return Version Number as a string.
     */
    static std::string getVersion();

    /**
     * Obtain the unique fingerprint for the current device.
     * The fingerprint is used to generate a hardware locked token.
     *
     * @return The fingerprint as a string.
     */

    static std::string getDeviceFingerprint();

    /**
     * Obtain the GPU UUID(s).
     * The GPU UUID(s) are used to generate a GPU locked token.
     * Note, this function can only be run on the GPU SDK, otherwise it will throw an exception.
     *
     * @return The GPU UUID(s).
     */
    static std::string getGPUUUID();

    /**
     *  Preprocess the image to be used by the other methods.
     *  Creates a copy of the image buffer.
     *
     * @param[in] image An 8-bit decoded image array, in the CPU memory or the GPU memory.
     * @param[in] width The image width.
     * @param[in] height The image height.
     * @param[in] color The image color model, see ColorCode.
     * @param[out] tfImage The preprocesses image.
     * @param[in] stride The distance between image array rows in bytes, also known as step size
     * (used only for pointers to the GPU memory).
     *
     * @return Error code, see ErrorCode.
     *
     */
    ErrorCode preprocessImage(uint8_t *image, uint16_t width, uint16_t height, ColorCode color,
                              TFImage &tfImage, int stride = 0);

    /**
     * Preprocess the image to be used by the other methods.
     *
     * @param[in] encodedImageBuffer Encoded image buffer, supports JPG, PNG, BMP, and TIFF.
     * @param[out] tfImage The preprocesses image.
     *
     * @return Error code, see ErrorCode.
     *
     */
    ErrorCode preprocessImage(const std::vector<uint8_t> &encodedImageBuffer, TFImage &tfImage);

    /**
     * Preprocess the image to be used by the other methods.
     *
     * @param[in] imageFile The path of a JPEG, JPG, PNG, BMP, or TIFF file (path must
     * not contain ~ ).
     * @param[out] tfImage The preprocesses image.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode preprocessImage(const std::string &imageFile, TFImage &tfImage);

    /**
     * Detect people and objects in the image.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[out] boundingBoxes A vector of BoundingBox objects that
     *             gets populated with the detected objects.
     *
     * @return Error code, see ErrorCode.
     *
     * @sa getObjectLabelString(ObjectLabel label)
     */
    ErrorCode detectObjects(const TFImage &tfImage, std::vector<BoundingBox> &boundingBoxes);

    /**
     * Draw object detector bounding boxes and labels on the image.
     * Note, you should not run any inference functions on the image after calling this method.
     *
     * @param[in] tfImage The input image on which to draw the labels.
     * @param[in] boundingBoxes The bounding boxes returned by detectObjects().
     * @param[in] scale Scale factor for the text size and bounding box thickness. Increase scale
     * factor for higher resolution images.
     * @return Error code, see ErrorCode.
     */
    ErrorCode drawObjectLabels(const TFImage &tfImage,
                               const std::vector<BoundingBox> &boundingBoxes,
                               unsigned int scale = 2);

    /**
     *  Converts an object label to a string. This can be used to
     *  display the object labels.
     *
     * @param[in] label The label of a BoundingBox.
     * @return The object label as a string.
     */
    static std::string getObjectLabelString(ObjectLabel label);

    /**
     * Detect the largest face in the image by area.
     * Use the \ref ConfigurationOptions.fdFilter to filter the detected faces.
     *
     * Refer to our <a
     * href="faq.html#what-is-the-impact-of-various-factors-on-the-face-recognition-similarity-score">FAQ
     * page</a> to understand the impact of face height on similarity score.
     *
     * The face detector is able to detect faces in the following dynamic height range.
     * Smallest detectable face = ((the larger of your image dimensions) / 640 * 20) pixels.
     * Largest detectable face = (your image height) pixels.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[out] faceBoxAndLandmarks The FaceBoxAndLandmarks containing the landmarks and bounding
     * box of the largest detected face in the image.
     * @param[out] found Whether a face was found in the image.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectLargestFace(const TFImage &tfImage, FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                bool &found);

    /**
     * Detect all the faces in the image.
     * Use the \ref ConfigurationOptions.fdFilter to filter the detected faces.
     *
     * Refer to our <a
     * href="faq.html#what-is-the-impact-of-various-factors-on-the-face-recognition-similarity-score">FAQ
     * page</a> to understand the impact of face height on similarity score.
     *
     * The face detector is able to detect faces in the following dynamic height range.
     * Smallest detectable face = ((the larger of your image dimensions) / 640 * 20) pixels.
     * Largest detectable face = (your image height) pixels.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[out] faceBoxAndLandmarks A vector of \ref FaceBoxAndLandmarks representing each of the
     * detected faces. If not faces are found, the vector will be empty. The detected faces are
     * sorted in order of descending face score.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectFaces(const TFImage &tfImage,
                          std::vector<FaceBoxAndLandmarks> &faceBoxAndLandmarks);

    /**
     * Draw face detection bounding box and 5 keypoint landmarks.
     * Note, you should not run any inference functions on the image after calling this method.
     *
     * @param[in] tfImage The input image on which to draw the labels.
     * @param[in] faceBoxAndLandmarks The face to annotate.
     * @param[in] includeLandmarks Indicates if the 5 keypoint landmarks should also be drawn.
     * @param[in] color The color to use for the annotation.
     * @param[in] scale Scale factor for the bounding box thickness. Increase scale factor for
     * higher resolution images.
     * @return Error code, see ErrorCode.
     */
    ErrorCode drawFaceBoxAndLandmarks(const TFImage &tfImage,
                                      const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                      bool includeLandmarks = false,
                                      const ColorRGB &color = ColorRGB(0, 255, 255),
                                      unsigned int scale = 2);

    /**
     * Obtain the 106 face landmarks.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or
     * detectLargestFace().
     * @param[out] landmarks An array of 106 face landmark points.
     *
     * @return Error code, see ErrorCode.
     */

    ErrorCode getFaceLandmarks(const TFImage &tfImage,
                               const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                               Landmarks &landmarks);

    /**
     * Obtain the 106 face landmarks.
     * his batch processing method increases throughput when using GPU inference.T
     *
     * @param[in] tfImages A vector of input images.
     * @param[in] faceBoxAndLandmarksVec A vector of the FaceBoxAndLandmarks.
     * @param[out] landmarksVec A vector of arrays representing the 106 face landmark points in each
     * face image.
     * @return Error code, see ErrorCode.
     */
    ErrorCode getFaceLandmarks(const std::vector<TFImage> &tfImages,
                               const std::vector<FaceBoxAndLandmarks> &faceBoxAndLandmarksVec,
                               std::vector<Landmarks> &landmarksVec);

    /**
     * Draw the 106 face landmarks on the face image.
     * Note, you should not run any inference functions on the image after calling this method.
     *
     * @param[in] tfImage The input image on which to draw the landmarks.
     * @param[in] landmarks The 106 face landmarks.
     * @param[in] color The color to use for the annotation.
     * @param[in] scale Size of the landmark points draw on the face. Increase scale factor for
     * higher resolution images.
     *
     * @return error code
     */
    ErrorCode drawFaceLandmarks(const TFImage &tfImage, const Landmarks &landmarks,
                                const ColorRGB &color = ColorRGB(0, 255, 255),
                                unsigned int scale = 2);

    /**
     *  Align the the detected face to be optimized for passing to
     *  feature extraction. If using the face chip with Trueface algorithms (ex face recognition),
     *  do not change the default margin and scale values.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks The FaceBoxAndLandmarks returned
     *            by detectLargestFace() or detectFaces().
     * @param[out] tfFacechip A TFFacechip to store the output facechip.
     * @param[in] marginLeft Adds a margin to the left side of the face chip.
     * @param[in] marginTop Adds a margin to the top side of the face chip.
     * @param[in] marginRight Adds a margin to the right side of the face chip.
     * @param[in] marginBottom Adds a margin to the bottom side of the face chip.
     * @param[in] scale Changes the scale of the face chip.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode extractAlignedFace(const TFImage &tfImage,
                                 const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                 TFFacechip &tfFacechip, int marginLeft = 0, int marginTop = 0,
                                 int marginRight = 0, int marginBottom = 0, float scale = 1.0);

    /**
     * Detect the largest face in the image by area and return its feature
     * vector (Faceprint).
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[out] faceprint A Faceprint object which will contain
     *             the face feature vector.
     * @param[out] foundFace Indicates if a face was detected in the image.
     * If no face was detected, then the faceprint will be empty.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode getLargestFaceFeatureVector(const TFImage &tfImage, Faceprint &faceprint,
                                          bool &foundFace);

    /**
     * Extract the face feature vector (Faceprint) from an aligned face chip.
     *
     * @param[in] tfFacechip TFFacechip returned by extractAlignedFace().
     * Face image must be have size of 112x112 pixels (default extractAlignedFace() margin and scale
     * values).
     * @param[out] faceprint A Faceprint object which will contain
     *             the face feature vector.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode getFaceFeatureVector(const TFFacechip &tfFacechip, Faceprint &faceprint);

    /**
     * Extract the face feature vectors (Faceprints) from the face chips.
     * This batch processing method increases throughput when using GPU inference.
     *
     * @param[in] tfFacechips Vector of TFFacechip.
     * @param[out] faceprints Vector of Faceprint object which will contain
     *             the face feature vectors.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode getFaceFeatureVectors(const std::vector<TFFacechip> &tfFacechips,
                                    std::vector<Faceprint> &faceprints);

    /**
     * Extract the face feature vector (Faceprint) from the face box.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks The FaceBoxAndLandmarks returned by detectFaces() or
     * detectLargestFace().
     * @param[out] faceprint Faceprint to contain the face template.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode getFaceFeatureVector(const TFImage &tfImage,
                                   const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                   Faceprint &faceprint);

    /**
     * Convert a Faceprint into a json string.
     *
     * @param[in] faceprint The Faceprint to stringify.
     *
     * @return The json string representation of the Faceprint.
     */
    static std::string faceprintToJson(const Faceprint &faceprint);

    /**
     * Populate a Faceprint from a json string.
     *
     * @param[in] jsonStr The json string representation of a Faceprint, generated from the
     * faceprintToJson() function.
     * @param[out] faceprint The Faceprint which will be populated from the json string.
     *
     * @return Error code, see ErrorCode.
     */
    static ErrorCode jsonToFaceprint(const std::string &jsonStr, Faceprint &faceprint);

    /**
     * Compute the similarity between two feature vectors, or how similar two faces are.
     * Note, while the matchProbability may be more intuitive to understand, match thresholding
     * should be performed on the similarityMeasure. To understand the difference between these two
     * metrics, refer to our <a
     * href="faq.html#what-is-the-difference-between-similarity-score-and-match-probability">FAQ
     * page</a>. Refer to our <a href="https://docs.trueface.ai/roc-curves">ROC curves</a> when
     * selecting a threshold.
     *
     * @param[in] faceprint1 The first Faceprint to be compared.
     * @param[in] faceprint2 The second Faceprint to be compared.
     * @param[out] matchProbability The probability the two face feature vectors are a match.
     * @param[out] similarityMeasure The computed similarity measure.
     *
     * @return Error code, see ErrorCode.
     *
     * @sa getLargestFaceFeatureVector
     */
    static ErrorCode getSimilarity(const Faceprint &faceprint1, const Faceprint &faceprint2,
                                   float &matchProbability, float &similarityMeasure);

    /**
     * Estimate the head orientation.
     * Refer to our <a
     * href="faq.html#what-is-the-impact-of-various-factors-on-the-face-recognition-similarity-score">FAQ
     * page</a> to understand the impact of face orientation on similarity score.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks The face box and landmarks, returned by detectLargestFace().
     * @param[in] landmarks The detailed face landmarks, returned by getFaceLandmarks().
     * @param[out] headOrientation The predicted head orientation.
     *
     * @return Error code, see ErrorCode.
     * */

    ErrorCode estimateHeadOrientation(const TFImage &tfImage,
                                      const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                      const Landmarks &landmarks, HeadOrientation &headOrientation);

    /**
     * Draw the 3D face orientation axes.
     * Note, you should not run any inference functions on the image after calling this method.
     *
     * @param[in] tfImage The input image on which to draw the rotation axes.
     * @param[in] faceBoxAndLandmarks The face for which the orientation axes will be drawn.
     * @param[in] headOrientation The head orientation, returned by estimateHeadOrientation().
     * @param[in] scale The scale factor for the axis line thickness and length. Increase scale
     * factor for higher resolution images.
     * @return Error code.
     */
    ErrorCode drawHeadOrientationAxes(const TFImage &tfImage,
                                      const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                      const HeadOrientation &headOrientation,
                                      unsigned int scale = 2);

    /**
     * Draw 3D box on the face.
     * Note, you should not run any inference functions on the image after calling this method.
     * @param[in] tfImage The input image on which to draw the rotation axes.
     * @param[in] headOrientation The head orientation, returned by estimateHeadOrientation().
     * @param[in] scale The scale factor for the box line thickness. Increase scale factor for
     * higher resolution images.
     * @return Error code.
     */

    ErrorCode drawHeadOrientationBox(const TFImage &tfImage, const HeadOrientation &headOrientation,
                                     unsigned int scale = 2);

    /**
     * Estimate the quality of the face image for recognition, with a score of 1 indicating perfect
     * quality. This method currently calls detectFaceImageBlur() behind the scene and converts the
     * score to either 1 or 0. In the future, more metrics will be added internally in this function
     *
     * @param[in] tfFacechip The TFFacechip returned by extractAlignedFace().
     * @param[out] quality A value between 0 to 1.
     * We suggest using a threshold of 0.5 as a filter for enrollment images.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode estimateFaceImageQuality(const TFFacechip &tfFacechip, float &quality);

    /**
     * Template quality score is a metric defined to measure the biometric utility or effectiveness
     * of a face sample for the purpose of face recognition. It does not measure the visual quality
     * of the sample. Biometric utility is specifically about the sample's effectiveness in a
     * biometric system, which may not always align with human perception of image quality. For
     * instance, a visually clear image might have low biometric utility if the facial features are
     * not captured in a way conducive to accurate matching.
     *
     * @param[in] tfFacechip The TFFacechip returned by extractAlignedFace().
     * @param[out] isTemplateQualityGood Indicates if the face chip is suitable for face
     * recognition.
     * @param[out] score The quality score. Can be used for custom thresholding (default > 1.f is
     * good quality).
     * @return Error code, see ErrorCode.
     */
    ErrorCode estimateFaceTemplateQuality(const TFFacechip &tfFacechip, bool &isTemplateQualityGood,
                                          float &score);

    /**
     * Template quality score is a metric defined to measure the biometric utility or effectiveness
     * of a face sample for the purpose of face recognition. It does not measure the visual quality
     * of the sample. Biometric utility is specifically about the sample's effectiveness in a
     * biometric system, which may not always align with human perception of image quality. For
     * instance, a visually clear image might have low biometric utility if the facial features are
     * not captured in a way conducive to accurate matching. This batch processing method increases
     * throughput when using GPU inference.
     *
     * @param[in] tfFacechips A vector of TFFacechip
     * @param[out] areTemplateQualitiesGood indicates if the face chips are suitable for face
     * recognition.
     * @param[out] scores The quality scores. Can be used for custom thresholding (default > 1.f is
     * good quality).
     * @return Error code, see ErrorCode.
     */
    ErrorCode estimateFaceTemplateQualities(const std::vector<TFFacechip> &tfFacechips,
                                            std::vector<bool> &areTemplateQualitiesGood,
                                            std::vector<float> &scores);

    /**
     * Determine if the image is over or under exposed, indicating suboptimal lighting conditions
     * for face recognition. Refer to our <a
     * href="faq.html#what-is-the-impact-of-various-factors-on-the-face-recognition-similarity-score">FAQ
     * page</a> to understand the impact of image exposure on similarity score.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks The FaceBoxAndLandmarks of the face to analyze.
     * @param[out] result The FaceImageQuality output.
     * @param[out] percentImageBright The percentage of pixels in the image which are above the
     * brightness threshold.
     * @param[out] percentImageDark The percentage of pixels in the image which are below the
     * darkness threshold.
     * @param[out] percentFaceBright The percentage of pixels in the face region which are above the
     * brightness threshold.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode checkFaceImageExposure(const TFImage &tfImage,
                                     const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                     FaceImageQuality &result, float &percentImageBright,
                                     float &percentImageDark, float &percentFaceBright);

    /**
     * Determine if the eyes of the specified face are open or closed.
     * Note, this function does not work well with eyeglasses, which can be detected using
     * `Trueface::SDK::detectGlasses`.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] targetFace The face on which to run blink detection.
     * @param[out] blinkState The output of the blink prediction.
     *
     * @return Error code, see ErrorCode. Note, if the face is not frontal facing, an error will be
     * returned.
     */
    ErrorCode detectBlink(const TFImage &tfImage, const FaceBoxAndLandmarks &targetFace,
                          BlinkState &blinkState);

    /**
     * Determine if the eyes of the specified faces are open or closed.
     * This batch processing method increases throughput when using GPU inference.
     *
     * @param[in] tfImages A list of input images.
     * @param[in] keypointsVec A list of the face landmarks for the faces on which to run blink
     * detection.
     * @param[out] blinkStates The output list of blink prediction results.
     * @return Error code, see ErrorCode. Note, if one of the face is not frontal facing, an error
     * will be returned.
     */
    ErrorCode detectBlinks(const std::vector<TFImage> &tfImages,
                           const std::vector<Landmarks> &keypointsVec,
                           std::vector<BlinkState> &blinkStates);

    /**
     * Ensures that the face size meets the requirements for active spoof.
     * Must check return value of function!
     * Active spoof works by analyzing the way a persons face changes as they move closer to a
     camera.
     * The active spoof solution therefore expects the face a certain distance from the camera.
     * In the far image, the face should be about 18 inches from the camera, while
     * in the near image, the face should be 7-8 inches from the camera.
     * This function must be called before calling detectActiveSpoof().
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks The face on which to run active spoof detection.
     * @param[in] activeSpoofStage The stage of the image, either near stage or far stage.

     * @return Error code, see ErrorCode.
     * If `ErrorCode::NO_ERROR` is returned, then the image is eligible for active spoof detection.
     * If `ErrorCode::FACE_TOO_CLOSE` or `ErrorCode::FACE_TOO_FAR` is returned, the image is not
     eligible for active spoof detection.
     *
     */
    ErrorCode checkSpoofImageFaceSize(const TFImage &tfImage,
                                      const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                      ActiveSpoofStage activeSpoofStage);

    /**
     *
     * Detect if there is a presentation attack attempt.
     * Must call checkSpoofImageFaceSize() on both input faces before calling this function.
     *
     * @param[in] nearFaceLandmarks The face landmarks of the near face, obtained by calling
     * getFaceLandmarks().
     * @param[in] farFaceLandmarks The face landmarks of the far face, obtained by calling
     * getFaceLandmarks().
     * @param[out] spoofScore The output spoof score.
     * If the spoof score is above the threshold, then it is classified as a real face.
     * If the spoof score is below the threshold, then it is classified as a fake face.
     * @param[out] spoofPrediction The predicted spoof result, using a spoofScore threshold of 1.05.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectActiveSpoof(const Landmarks &nearFaceLandmarks,
                                const Landmarks &farFaceLandmarks, float &spoofScore,
                                SpoofLabel &spoofPrediction);

    /**
     * Detect if there is a presentation attack attempt.
     * Passive spoof detection only works with "selfie-style", meaning the face must be up close and
     * take up most of the frame.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or
     * detectLargestFace().
     * @param[out] result The SpoofLabel results for the face image.
     * @param[out] spoofScore The probability that the image is a spoof attempt
     * (1 indicates real image, 0 indicates spoof attempt).
     * @param[in] threshold The spoof score threshold above which it is considered a real image.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectSpoof(const TFImage &tfImage, const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                          SpoofLabel &result, float &spoofScore, float threshold = 0.75);

    /**
     * Detect if there is a presentation attack attempt.
     * Passive spoof detection only works with "selfie-style", meaning the face must be up
     * close and take up most of the frame.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or
     * detectLargestFace().
     * @param[in] headOrientation The head orientation, returned by estimateHeadOrientation().
     * @param[in] blinkState The output of the blink prediction, as returned by
     * detectBlink()
     * @param[in] maskLabel The predicted MaskLabel for face image, as returned by
     * detectMask()
     * @param[out] result The SpoofLabel results for the face image.
     * @param[out] spoofScore The probability that the image is a spoof attempt
     * (1 indicates real image, 0 indicates spoof attempt).
     * @param[in] threshold The spoof score threshold above which it is considered a real
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectSpoof(const TFImage &tfImage, const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                          const HeadOrientation &headOrientation, const BlinkState &blinkState,
                          const MaskLabel &maskLabel, SpoofLabel &result, float &spoofScore,
                          float threshold = 0.75);
    /**
     *
     * Create a connection to a new or existing database.
     *
     * If the database does not exist, a new one will be created with the provided name.
     *
     * If the `Trueface::DatabaseManagementSystem::NONE` (memory only)
     * configuration option is selected, this function does not need to be called (and is a harmless
     * no-op).
     *
     * If connecting to a hosted PostgreSQL database (AWS, Digital Ocean, etc), see the
     * `HOSTED_DATABASE` environment variable on the <a href="./environment.html">environment
     * variable page</a>.
     *
     * If connecting to a PostgreSQL database, use the "host" parameter to
     * specify the host name url (ex. mydatabase.us-east-1.rds.amazonaws.com, localhost, etc).
     *
     * If connecting to a PostgreSQL database and you only have the numeric IP address of the
     * server, then you must use the "hostaddr" parameter. Using the "hostaddr" parameter
     * instead of "host" will avoid a host name look-up.
     *
     * Note, the database name cannot contain uppercase characters.
     *
     * @param[in] databaseConnectionString
     * If `Trueface::DatabaseManagementSystem::SQLITE` is selected, this should be the filepath to
     * the database. ex. "/myPath/myDatabase.db".
     * If `Trueface::DatabaseManagementSystem::POSTGRESQL` is selected, this should be a database
     * connection string.
     * <a
     * href="https://www.postgresql.org/docs/current/libpq-connect.html#LIBPQ-PARAMKEYWORDS">Here</a>
     * is a list of all supported PostgreSQL connection parameters.
     * ex. "hostaddr=192.168.1.0 port=5432 dbname=face_recognition user=postgres
     * password=my_password". ex. "host=mydatabase.us-east-1.rds.amazonaws.com port=5432
     * dbname=face_recognition user=postgres password=m_password". To enable ssl, add
     * "sslmode=require" to the connection string.
     *
     * @return Error code, see ErrorCode.
     */

    ErrorCode createDatabaseConnection(const std::string &databaseConnectionString);

    /**
     *
     * Create a new collection, or load data from an existing collection into memory (RAM) if one
     * with the provided name already exists in the database. Will unload any collections previously
     * loaded into memory. Equivalent to calling createCollection() then loadCollection().
     * Collection name cannot be a reserved <a
     * href="https://www.postgresql.org/docs/current/sql-keywords-appendix.html">key word</a>.
     *
     * @param[in] collectionName The name of the collection.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode createLoadCollection(const std::string &collectionName);

    /**
     *
     * Create a new collection in the database.
     * Before enrolling Faceprints into newly created collection, must call loadCollection() or
     * loadCollections(). If a collection with the provided name already exists, this is a harmless
     * no-op. Collection name cannot be a reserved <a
     * href="https://www.postgresql.org/docs/current/sql-keywords-appendix.html">key word</a>.
     *
     * @param[in] collectionName The name of the collection.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode createCollection(const std::string &collectionName);

    /**
     *
     * Loads the specified collection into memory, while unloading any collections previously loaded
     * into memory. Must be called before enrolling Faceprints or calling identification functions.
     * Refer to our <a
     * href="faq.html#what-is-the-difference-between-loadcollection-and-loadcollectionpersist">FAQ
     * page</a> for more information on this method.
     *
     * @param[in] collectionName The name of the collection to load into memory.
     *
     * @return Error code, see ErrorCode.
     * Will return an error if the collection does not exist.
     */
    ErrorCode loadCollection(const std::string &collectionName);

    /**
     *
     * Loads the specified collection into memory. Does not unload any collections previously loaded
     * into memory. Must be called before enrolling Faceprints or calling identification functions.
     * Refer to our <a
     * href="faq.html#what-is-the-difference-between-loadcollection-and-loadcollectionpersist">FAQ
     * page</a> for more information on this method.
     *
     * @param[in] collectionName The name of the collection to load into memory.
     *
     * @return Error code, see ErrorCode.
     * Will return an error if the collection does not exist.
     */
    ErrorCode loadCollectionPersist(const std::string &collectionName);

    /**
     * Load the specified collections into memory, while unloading any collections previously loaded
     * into memory. Must be called before enrolling Faceprints or calling identification functions.
     *
     * @param[in] collectionNames The name of the collections to load into memory.
     * @return Error code, see ErrorCode.
     * Will return an error if one or more of the specified collections does not exist.
     *
     */
    ErrorCode loadCollections(const std::vector<std::string> &collectionNames);

    /**
     *
     * Deletes a collection from the current database.
     *
     * @param[in] collectionName The name of the collection to delete.
     *
     * @return Error code, see ErrorCode.
     * Will return an error if the collection does not exist.
     */
    ErrorCode deleteCollection(const std::string &collectionName);

    /**
     * Enroll a Faceprint for a new or existing identity in the collection.
     *
     * @param[in] faceprint The Faceprint to enroll in the collection.
     * @param[in] identity The identity corresponding to the Faceprint.
     * @param[out] UUID Universally unique identifier corresponding to the Faceprint.
     * @param[in] collectionName The name of the collection in which to enroll the Faceprint.
     * If only a single collection has been loaded into memory, this can remain an empty string (in
     * order to maintain backwards compatibility). If multiple collections are loaded in memory,
     * then the collection name parameter must be specified.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode enrollFaceprint(const Faceprint &faceprint, const std::string &identity,
                              std::string &UUID, const std::string &collectionName = "");

    /**
     * Get a list of the names of all the collections in the database.
     * Collection names can then be passed to getCollectionMetadata() and getCollectionIdentities().
     *
     * @param[out] collectionNames List of collection names in the database.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode getCollectionNames(std::vector<std::string> &collectionNames);

    /**
     * Returns a list of collections which are currently loaded in memory.
     * @param[out] collectionNames List of collections in memory.
     * @return Error code, see ErrorCode.
     */
    ErrorCode getLoadedCollectionNames(std::vector<std::string> &collectionNames);

    /**
     * Get the metadata for the specified collection in the database, loaded or unloaded.
     *
     * @param[in] collectionName The name of the collection for which to retrieve the metadata.
     * @param[out] metadata The metadata for the specified collection.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode getCollectionMetadata(const std::string &collectionName,
                                    CollectionMetadata &metadata);

    /**
     * Get a map of identities and UUIDs for the specified collection in the database, loaded or
     * unloaded. This can be a slow operation (especially for unloaded collections), call sparingly.
     *
     * @param[in] collectionName The name of the collection for which to retrieve the identities.
     * @param[out] identities A map of identities and corresponding UUIDs.
     * There can be multiple UUIDs mapped to a single identity if multiple Faceprints were enrolled
     * for that identity.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode
    getCollectionIdentities(const std::string &collectionName,
                            std::unordered_multimap<std::string, std::string> &identities);

    /**
     * Remove a Faceprint from the collection using the UUID.
     *
     * @param[in] UUID The universally unique identifier returned by enrollFaceprint().
     * @param[in] collectionName The name of the collection from which to remove the Faceprint.
     * If only a single collection has been loaded into memory, this can remain an empty string (in
     * order to maintain backwards compatibility). If multiple collections are loaded in memory,
     * then the collection name parameter must be specified.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode removeByUUID(const std::string &UUID, const std::string &collectionName = "");

    /**
     * Remove all Faceprints in the collection corresponding to the identity.
     *
     * @param[in] identity The identity to remove from the collection.
     * @param[out] numFaceprintsRemoved The the number of Faceprint which were removed for that
     * identity.
     * @param[in] collectionName The name of the collection from which to remove the Faceprint.
     * If only a single collection has been loaded into memory, this can remain an empty string (in
     * order to maintain backwards compatibility).
     * If multiple collections are loaded in memory, * then the collection name parameter must
     * be specified.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode removeByIdentity(const std::string &identity, unsigned int &numFaceprintsRemoved,
                               const std::string &collectionName = "");

    /**
     * Get the top match Candidate in the collection and the corresponding similarity score and
     * match probability.
     *
     * @param[in] faceprint The Faceprint to be identified.
     * @param[out] candidate The top match Candidate.
     * @param[out] found Set to true if a match is found.
     * @param[in] threshold The similarity score threshold above which it is considered a match.
     * Higher thresholds may result in faster queries.
     * Refer to our <a href="https://docs.trueface.ai/roc-curves">ROC curves</a> when selecting a
     * threshold.
     * @param[in] collectionName The name of the collection which should be searched.
     * If only a single collection has been loaded into memory, this can remain an empty string (in
     * order to maintain backwards compatibility).
     * If multiple collections are loaded in memory, then the collection name parameter must be
     * specified.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode identifyTopCandidate(const Faceprint &faceprint, Candidate &candidate, bool &found,
                                   float threshold = 0.4f, const std::string &collectionName = "");

    /**
     * Draw a bounding box around the face and a label containing the identity and match
     * probability. Note, you should not run any inference functions on the image after calling this
     * method.
     *
     * @param[in] tfImage The input image on which to draw the labels.
     * @param[in] faceBoxAndLandmarks The face to annotate.
     * @param[in] candidate The candidate information.
     * @param[in] color The color to use for the annotation.
     * @param[in] scale The scale factor for the text size and bounding box thickness. Increase
     * scale factor for higher resolution images.
     * @return ErrorCode.
     */
    ErrorCode drawCandidateBoundingBoxAndLabel(const TFImage &tfImage,
                                               const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                                               const Candidate &candidate,
                                               const ColorRGB &color = ColorRGB(0, 255, 0),
                                               unsigned int scale = 2);

    /**
     * Get the top match Candidate in the collection and the corresponding similarity score and
     * match probability for each probe Faceprint. Like identifyTopCandidate(), but runs search
     * queries in parallel and improves throughput.
     *
     * @param[in] faceprints A vector of Faceprint to be identified.
     * @param[out] candidates The top match Candidate for each probe Faceprint.
     * @param[out] found Set to true if a match is found for the probe Faceprint.
     * @param[in] threshold The similarity score threshold above which it is considered a match.
     * Higher thresholds may result in faster queries.
     * Refer to out <a href="https://docs.trueface.ai/roc-curves">ROC curves</a> when selecting a
     * threshold.
     * @param[in] collectionName The name of the collection which should be searched.
     * If only a single collection has been loaded into memory, this can remain an empty string (in
     * order to maintain backwards compatibility).
     * If multiple collections are loaded in memory, then the collection name parameter must
     * be specified.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode batchIdentifyTopCandidate(const std::vector<Faceprint> &faceprints,
                                        std::vector<Candidate> &candidates,
                                        std::vector<bool> &found, float threshold = 0.4f,
                                        const std::string &collectionName = "");

    /**
     * Get a list of the top n match Candidate in the collection and their corresponding similarity
     * scores and match probabilities. Ideal for investigation mode.
     *
     * @param[in] faceprint The Faceprint to be identified.
     * @param[out] candidates A list of the top n matches (or fewer) Candidate, ordered by
     * descending similarity score.
     * @param[out] found Set to true if at leaset one match is found.
     * @param[in] numCandidates Max number of Candidate identities to return.
     * @param[in] threshold The similarity score threshold above which it is considered a match.
     * Higher thresholds may result in faster queries.
     * Refer to out <a href="https://docs.trueface.ai/roc-curves">ROC curves</a> when selecting a
     * threshold.
     * @param[in] collectionName The name of the collection which should be searched.
     * If only a single collection has been loaded into memory, this can remain an empty string (in
     * order to maintain backwards compatibility).
     * If multiple collections are loaded in memory, then the collection name parameter must
     * be specified.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode identifyTopCandidates(const Faceprint &faceprint, std::vector<Candidate> &candidates,
                                    bool &found, unsigned int numCandidates, float threshold = 0.4f,
                                    const std::string &collectionName = "");

    /**
     * The PostgreSQL DatabaseManagementSystem has synchronization built-in, in which the database
     * will propagate any notifications to to any of the connected clients.
     *
     * Even if the network connection is interrupted, any notifications which were sent out during
     * this period will be received when the network connection is reestablished, ensuring the
     * client collection remains synchronized.
     *
     * Under rare circumstances, the connection to the database may be broken
     * entirely. If any notifications are sent out during this period, the client will not know
     * about them, and will still not know about them when database connection is reestablished.
     *
     * In this rare occurrence, it is possible for a client collection to go out of sync with the
     * database (only if another client made any collection modifications during the time when the
     * connection was broken).
     *
     * Therefore, you can choose to register a callback method which is called any time the
     * connection to the database is broken. We advise capturing a flag or conditional variable by
     * reference and setting it within the method you define, and then check for the flag in your
     * application and respond appropriately.
     *
     * You could for example then choose to reload the collection into memory, ensuring that it is
     * fully up to date with the database in-case a database notification was missed. If you do
     * choose to register a callback method, then when a disconnect is detected, the callback method
     * will be called, and the SDK will not try to reconnect to the database.
     *
     * @param[in] callback The callback method which is called anytime the database is disconnected.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode registerDatabaseDisconnectionCallback(std::function<void()> callback);

    /**
     * Detect whether the face in the image is wearing a mask or not.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or
     * detectLargestFace().
     * @param[out] result The predicted MaskLabel for face image.
     * @param[out] maskScore The mask score for this image. This can be used for setting custom
     * thresholds that work better for the use case. By default, we use a mask score greater than
     * 0.82 to determine that no mask was detected.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectMask(const TFImage &tfImage, const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                         MaskLabel &result, float &maskScore);

    /**
     * Detect whether the faces are wearing a mask or not.
     * This batch processing method increases throughput when using GPU inference.
     *
     * @param[in] tfFacechips A vector of TFFacechip on which to run mask detection.
     * @param[out] results The predicted MaskLabel for the input face chips.
     * @param[out] maskScores The mask scores for these images. This can be used for setting custom
     * thresholds that work better for the use case. By default, we use a mask score greater than
     * 0.82 to determine that no mask was detected.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectMasks(const std::vector<TFFacechip> &tfFacechips,
                          std::vector<MaskLabel> &results, std::vector<float> &maskScores);

    /**
     * Detect whether the face in the image is wearing any type of eye glasses or not.
     *
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[in] faceBoxAndLandmarks FaceBoxAndLandmarks returned by detectFaces() or
     * detectLargestFace().
     * @param[out] result The predicted GlassesLabel for face image.
     * @param[out] glassesScore The glasses score for this image. This can be used for setting
     * custom thresholds that work better for the use case. By default, we use a glasses score
     * greater than 0.0 to determine that glasses were detected.
     *
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectGlasses(const TFImage &tfImage, const FaceBoxAndLandmarks &faceBoxAndLandmarks,
                            GlassesLabel &result, float &glassesScore);

    /**
     * Detect the orientation of a face image. Returns the rotation required to achieve neutral
     * orientation. As this method does add overhead, we advise only adding it to your pipeline when
     * doing offline processing (ex. reading a database of ID images where some ID images may not be
     * oriented correctly). When consuming a video stream, this method is not necessary as you can
     * guarantee all the frames will have the same orientation.
     * @param[in] tfImage The input image returned by preprocessImage().
     * @param[out] imageRotation The rotation required in order to achieve neutral orientation.
     * This can be passed directly to the Image::rotate() method in order to correctly rotate the
     * image.
     * @return Error code, see ErrorCode.
     */
    ErrorCode getFaceImageRotation(const TFImage &tfImage, RotateFlags &imageRotation);

    /**
     * Detect the orientation of a list of face images. Returns the rotation required to achieve
     * neutral orientation. As this method does add overhead, we advise only adding it to your
     * pipeline when doing offline processing (ex. reading a database of ID images where some ID
     * images may not be oriented correctly). When consuming a video stream, this method is not
     * necessary as you can guarantee all the frames will have the same orientation. This batch
     * processing method increases throughput when using GPU inference.
     * @param[in] tfImages The list of input images.
     * @param[out] imageRotations The rotations required in order to achieve neutral orientation.
     * These orientations can be passed directly to the Image::rotate() method in order to correctly
     * rotate the images.
     * @return Error code, see ErrorCode.
     */
    ErrorCode getFaceImageRotations(const std::vector<TFImage> &tfImages,
                                    std::vector<RotateFlags> &imageRotations);

    /**
     * Determine if the face image is blurry or is of good quality for face recognition.
     * Refer to our <a
     * href="faq.html#what-is-the-impact-of-various-factors-on-the-face-recognition-similarity-score">FAQ
     * page</a> to understand the impact of face blur on similarity score.
     * @param[in] tfFacechip The TFFacechip returned by extractAlignedFace().
     * @param[out] result The FaceImageQuality output.
     * @param[out] score The probability that the face is blurry.
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectFaceImageBlur(const TFFacechip &tfFacechip, FaceImageQuality &result,
                                  float &score);

    /**
     * Determine if the provided face images are blurry or are of good quality for face recognition.
     * This batch processing method increases throughput when using GPU inference.
     * Refer to our <a
     * href="faq.html#what-is-the-impact-of-various-factors-on-the-face-recognition-similarity-score">FAQ
     * page</a> to understand the impact of face blur on similarity score.
     *
     * @param[in] tfFacechips List of input TFFacechip.
     * @param[out] results The FaceImageQuality outputs.
     * @param[out] scores List of probabilities indicating if the faces are blurry.
     * @return Error code, see ErrorCode.
     */
    ErrorCode detectFaceImageBlurs(const std::vector<TFFacechip> &tfFacechips,
                                   std::vector<FaceImageQuality> &results,
                                   std::vector<float> &scores);

private:
    class Impl;
    std::unique_ptr<Impl> pImpl;
    friend class Internal;
    friend class Image;
};

} // namespace Trueface
